Class: Reline::ANSI
- Inherits:
-
IO
- Object
- IO
- Reline::ANSI
- Defined in:
- lib/reline/io/ansi.rb
Constant Summary collapse
- ANSI_CURSOR_KEY_BINDINGS =
{ # Up 'A' => [:ed_prev_history, {}], # Down 'B' => [:ed_next_history, {}], # Right 'C' => [:ed_next_char, { ctrl: :em_next_word, meta: :em_next_word }], # Left 'D' => [:ed_prev_char, { ctrl: :ed_prev_word, meta: :ed_prev_word }], # End 'F' => [:ed_move_to_end, {}], # Home 'H' => [:ed_move_to_beg, {}], }
- START_BRACKETED_PASTE =
String.new("\e[200~", encoding: Encoding::ASCII_8BIT)
- END_BRACKETED_PASTE =
String.new("\e[201~", encoding: Encoding::ASCII_8BIT)
Instance Attribute Summary collapse
-
#input ⇒ Object
writeonly
Sets the attribute input.
-
#output ⇒ Object
writeonly
Sets the attribute output.
Instance Method Summary collapse
- #both_tty? ⇒ Boolean
- #buffered_output ⇒ Object
- #clear_screen ⇒ Object
- #cursor_pos ⇒ Object
- #deprep(otio) ⇒ Object
- #empty_buffer? ⇒ Boolean
- #encoding ⇒ Object
- #erase_after_cursor ⇒ Object
- #get_screen_size ⇒ Object
-
#getc(timeout_second) ⇒ Object
if the usage expects to wait indefinitely, use Float::INFINITY for timeout_second.
- #hide_cursor ⇒ Object
- #in_pasting? ⇒ Boolean
-
#initialize ⇒ ANSI
constructor
A new instance of ANSI.
- #inner_getc(timeout_second) ⇒ Object
- #move_cursor_column(x) ⇒ Object
- #move_cursor_down(x) ⇒ Object
- #move_cursor_up(x) ⇒ Object
- #prep ⇒ Object
- #read_bracketed_paste ⇒ Object
- #read_single_char(timeout_second) ⇒ Object
-
#scroll_down(x) ⇒ Object
This only works when the cursor is at the bottom of the scroll range For more details, see https://github.com/ruby/reline/pull/577#issuecomment-1646679623.
- #set_bracketed_paste_key_bindings(config) ⇒ Object
- #set_default_key_bindings(config) ⇒ Object
- #set_default_key_bindings_ansi_cursor(config) ⇒ Object
- #set_default_key_bindings_comprehensive_list(config) ⇒ Object
- #set_screen_size(rows, columns) ⇒ Object
- #set_winch_handler(&handler) ⇒ Object
- #show_cursor ⇒ Object
- #ungetc(c) ⇒ Object
- #with_raw_input ⇒ Object
- #write(string) ⇒ Object
Constructor Details
#initialize ⇒ ANSI
Returns a new instance of ANSI.
22 23 24 25 26 27 28 |
# File 'lib/reline/io/ansi.rb', line 22 def initialize @input = STDIN @output = STDOUT @buf = [] @output_buffer = nil @old_winch_handler = nil end |
Instance Attribute Details
#input=(value) ⇒ Object (writeonly)
Sets the attribute input
20 21 22 |
# File 'lib/reline/io/ansi.rb', line 20 def input=(value) @input = value end |
#output=(value) ⇒ Object (writeonly)
Sets the attribute output
20 21 22 |
# File 'lib/reline/io/ansi.rb', line 20 def output=(value) @output = value end |
Instance Method Details
#both_tty? ⇒ Boolean
213 214 215 |
# File 'lib/reline/io/ansi.rb', line 213 def both_tty? @input.tty? && @output.tty? end |
#buffered_output ⇒ Object
225 226 227 228 229 230 231 |
# File 'lib/reline/io/ansi.rb', line 225 def buffered_output @output_buffer = +'' yield @output.write(@output_buffer) ensure @output_buffer = nil end |
#clear_screen ⇒ Object
273 274 275 276 |
# File 'lib/reline/io/ansi.rb', line 273 def clear_screen write "\e[2J" write "\e[1;1H" end |
#cursor_pos ⇒ Object
208 209 210 211 |
# File 'lib/reline/io/ansi.rb', line 208 def cursor_pos col, row = cursor_pos_internal(timeout: 0.5) if both_tty? Reline::CursorPos.new(col || 0, row || 0) end |
#deprep(otio) ⇒ Object
306 307 308 309 310 311 |
# File 'lib/reline/io/ansi.rb', line 306 def deprep(otio) # Disable bracketed paste write "\e[?2004l" if Reline.core.config.enable_bracketed_paste && both_tty? Signal.trap('WINCH', @old_winch_handler) if @old_winch_handler Signal.trap('CONT', @old_cont_handler) if @old_cont_handler end |
#empty_buffer? ⇒ Boolean
161 162 163 164 165 166 |
# File 'lib/reline/io/ansi.rb', line 161 def empty_buffer? unless @buf.empty? return false end !@input.wait_readable(0) end |
#encoding ⇒ Object
30 31 32 33 34 35 |
# File 'lib/reline/io/ansi.rb', line 30 def encoding @input.external_encoding || Encoding.default_external rescue IOError # STDIN.external_encoding raises IOError in Ruby <= 3.0 when STDIN is closed Encoding.default_external end |
#erase_after_cursor ⇒ Object
261 262 263 |
# File 'lib/reline/io/ansi.rb', line 261 def erase_after_cursor write "\e[K" end |
#get_screen_size ⇒ Object
172 173 174 175 176 177 178 179 180 |
# File 'lib/reline/io/ansi.rb', line 172 def get_screen_size s = @input.winsize return s if s[0] > 0 && s[1] > 0 s = [ENV["LINES"].to_i, ENV["COLUMNS"].to_i] return s if s[0] > 0 && s[1] > 0 [24, 80] rescue SystemCallError [24, 80] end |
#getc(timeout_second) ⇒ Object
if the usage expects to wait indefinitely, use Float::INFINITY for timeout_second
153 154 155 |
# File 'lib/reline/io/ansi.rb', line 153 def getc(timeout_second) inner_getc(timeout_second) end |
#hide_cursor ⇒ Object
253 254 255 |
# File 'lib/reline/io/ansi.rb', line 253 def hide_cursor write "\e[?25l" end |
#in_pasting? ⇒ Boolean
157 158 159 |
# File 'lib/reline/io/ansi.rb', line 157 def in_pasting? not empty_buffer? end |
#inner_getc(timeout_second) ⇒ Object
116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 |
# File 'lib/reline/io/ansi.rb', line 116 def inner_getc(timeout_second) while true # @buf may change while handling signal, so we need to check it in each loop iteration. return @buf.shift unless @buf.empty? break if @input.wait_readable(0.01) timeout_second -= 0.01 return nil if timeout_second <= 0 # handle_signal may call cursor_pos which modifies @buf Reline.core.line_editor.handle_signal end c = @input.getbyte # When "Escape non-ASCII Input with Control-V" is enabled in macOS Terminal.app, # all non-ascii bytes are automatically escaped with `C-v`. # "\xE3\x81\x82" (U+3042) becomes "\x16\xE3\x16\x81\x16\x82". (c == 0x16 && @input.tty? && @input.raw(min: 0, time: 0, &:getbyte)) || c rescue Errno::EIO # Maybe the I/O has been closed. nil end |
#move_cursor_column(x) ⇒ Object
233 234 235 |
# File 'lib/reline/io/ansi.rb', line 233 def move_cursor_column(x) write "\e[#{x + 1}G" end |
#move_cursor_down(x) ⇒ Object
245 246 247 248 249 250 251 |
# File 'lib/reline/io/ansi.rb', line 245 def move_cursor_down(x) if x > 0 write "\e[#{x}B" elsif x < 0 move_cursor_up(-x) end end |
#move_cursor_up(x) ⇒ Object
237 238 239 240 241 242 243 |
# File 'lib/reline/io/ansi.rb', line 237 def move_cursor_up(x) if x > 0 write "\e[#{x}A" elsif x < 0 move_cursor_down(-x) end end |
#prep ⇒ Object
300 301 302 303 304 |
# File 'lib/reline/io/ansi.rb', line 300 def prep # Enable bracketed paste write "\e[?2004h" if Reline.core.config.enable_bracketed_paste && both_tty? nil end |
#read_bracketed_paste ⇒ Object
141 142 143 144 145 146 147 148 149 150 |
# File 'lib/reline/io/ansi.rb', line 141 def read_bracketed_paste buffer = String.new(encoding: Encoding::ASCII_8BIT) until buffer.end_with?(END_BRACKETED_PASTE) c = inner_getc(Float::INFINITY) break unless c buffer << c end string = buffer.delete_suffix(END_BRACKETED_PASTE).force_encoding(encoding) string.valid_encoding? ? string : '' end |
#read_single_char(timeout_second) ⇒ Object
293 294 295 296 297 298 |
# File 'lib/reline/io/ansi.rb', line 293 def read_single_char(timeout_second) # Disable intr to read `C-c` `C-z` `C-\` for quoted insert @input.raw(intr: false) do super end end |
#scroll_down(x) ⇒ Object
This only works when the cursor is at the bottom of the scroll range For more details, see https://github.com/ruby/reline/pull/577#issuecomment-1646679623
267 268 269 270 271 |
# File 'lib/reline/io/ansi.rb', line 267 def scroll_down(x) return if x.zero? # We use `\n` instead of CSI + S because CSI + S would cause https://github.com/ruby/reline/issues/576 write "\n" * x end |
#set_bracketed_paste_key_bindings(config) ⇒ Object
56 57 58 59 60 |
# File 'lib/reline/io/ansi.rb', line 56 def set_bracketed_paste_key_bindings(config) [:emacs, :vi_insert, :vi_command].each do |keymap| config.add_default_key_binding_by_keymap(keymap, START_BRACKETED_PASTE.bytes, :bracketed_paste_start) end end |
#set_default_key_bindings(config) ⇒ Object
37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 |
# File 'lib/reline/io/ansi.rb', line 37 def set_default_key_bindings(config) set_bracketed_paste_key_bindings(config) set_default_key_bindings_ansi_cursor(config) set_default_key_bindings_comprehensive_list(config) { [27, 91, 90] => :completion_journey_up, # S-Tab }.each_pair do |key, func| config.add_default_key_binding_by_keymap(:emacs, key, func) config.add_default_key_binding_by_keymap(:vi_insert, key, func) end { # default bindings [27, 32] => :em_set_mark, # M-<space> [24, 24] => :em_exchange_mark, # C-x C-x }.each_pair do |key, func| config.add_default_key_binding_by_keymap(:emacs, key, func) end end |
#set_default_key_bindings_ansi_cursor(config) ⇒ Object
62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 |
# File 'lib/reline/io/ansi.rb', line 62 def set_default_key_bindings_ansi_cursor(config) ANSI_CURSOR_KEY_BINDINGS.each do |char, (default_func, modifiers)| bindings = [ ["\e[#{char}", default_func], # CSI + char ["\eO#{char}", default_func] # SS3 + char, application cursor key mode ] if modifiers[:ctrl] # CSI + ctrl_key_modifier + char bindings << ["\e[1;5#{char}", modifiers[:ctrl]] end if modifiers[:meta] # CSI + meta_key_modifier + char bindings << ["\e[1;3#{char}", modifiers[:meta]] # Meta(ESC) + CSI + char bindings << ["\e\e[#{char}", modifiers[:meta]] end bindings.each do |sequence, func| key = sequence.bytes config.add_default_key_binding_by_keymap(:emacs, key, func) config.add_default_key_binding_by_keymap(:vi_insert, key, func) config.add_default_key_binding_by_keymap(:vi_command, key, func) end end end |
#set_default_key_bindings_comprehensive_list(config) ⇒ Object
87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 |
# File 'lib/reline/io/ansi.rb', line 87 def set_default_key_bindings_comprehensive_list(config) { # xterm [27, 91, 51, 126] => :key_delete, # kdch1 [27, 91, 53, 126] => :ed_search_prev_history, # kpp [27, 91, 54, 126] => :ed_search_next_history, # knp # Console (80x25) [27, 91, 49, 126] => :ed_move_to_beg, # Home [27, 91, 52, 126] => :ed_move_to_end, # End # urxvt / exoterm [27, 91, 55, 126] => :ed_move_to_beg, # Home [27, 91, 56, 126] => :ed_move_to_end, # End }.each_pair do |key, func| config.add_default_key_binding_by_keymap(:emacs, key, func) config.add_default_key_binding_by_keymap(:vi_insert, key, func) config.add_default_key_binding_by_keymap(:vi_command, key, func) end end |
#set_screen_size(rows, columns) ⇒ Object
182 183 184 185 186 187 |
# File 'lib/reline/io/ansi.rb', line 182 def set_screen_size(rows, columns) @input.winsize = [rows, columns] self rescue SystemCallError self end |
#set_winch_handler(&handler) ⇒ Object
278 279 280 281 282 283 284 285 286 287 288 289 290 291 |
# File 'lib/reline/io/ansi.rb', line 278 def set_winch_handler(&handler) @old_winch_handler = Signal.trap('WINCH') do |arg| handler.call @old_winch_handler.call(arg) if @old_winch_handler.respond_to?(:call) end @old_cont_handler = Signal.trap('CONT') do |arg| @input.raw!(intr: true) if @input.tty? # Rerender the screen. Note that screen size might be changed while suspended. handler.call @old_cont_handler.call(arg) if @old_cont_handler.respond_to?(:call) end rescue ArgumentError # Signal.trap may raise an ArgumentError if the platform doesn't support the signal. end |
#show_cursor ⇒ Object
257 258 259 |
# File 'lib/reline/io/ansi.rb', line 257 def show_cursor write "\e[?25h" end |
#ungetc(c) ⇒ Object
168 169 170 |
# File 'lib/reline/io/ansi.rb', line 168 def ungetc(c) @buf.unshift(c) end |
#with_raw_input ⇒ Object
108 109 110 111 112 113 114 |
# File 'lib/reline/io/ansi.rb', line 108 def with_raw_input if @input.tty? @input.raw(intr: true) { yield } else yield end end |
#write(string) ⇒ Object
217 218 219 220 221 222 223 |
# File 'lib/reline/io/ansi.rb', line 217 def write(string) if @output_buffer @output_buffer << string else @output.write(string) end end |